from flask import request
from ..api.fields import (List, Nested)
from ..api.errors import ZeusNotFound


class QueryMixin(object):
    # 模型
    model = None
    # 排序
    order_by = None

    # 页码
    # type: int
    page = 1

    # 每页数据个数
    # type: int
    per_page = 20

    can_empty = None

    def get_paginate_args(self):
        """ 获取分页参数 """
        page = request.args.get('page', self.page, type=int) or self.page
        per_page = request.args.get('per_page', self.per_page, type=int) or self.per_page
        return page, per_page

    def get_model(self):
        """ 获取模型 """
        if not self.model:
            raise AttributeError('需要设置model值')
        return self.model

    def get_stmt(self):
        """ 自定义查询语句 """
        return None

    def get_query_filter(self):
        """ 自定义查询条件 """
        return {}

    def get_query(self, **kwargs):
        """ 获取查询语句 """
        if self.get_stmt():
            return self.get_stmt()

        model = self.get_model()

        stmt = model.query

        # 处理url主体,?之前
        filter_by_ = {}

        kwargs.update(self.get_query_filter())

        for k, v in kwargs.items():
            if self.model.has_property(k):
                filter_by_[k] = v

        if filter_by_:
            stmt = stmt.filter_by(**filter_by_)

        # 处理url参数?之后
        # 默认参数为多值,使用in操作符
        # 如果参数为单值,并且开头或结尾为%,使用like操作符
        filter_ = []

        for k, v in request.args.lists():
            if self.model.has_property(k):
                if len(v) == 1 and (v[0].startswith('%') or v[0].endswith('%')):
                    filter_.append(getattr(self.model, k).ilike(v[0]))
                else:
                    filter_.append(getattr(self.model, k).in_(v))

        if filter_:
            stmt = stmt.filter(*filter_)

        if self.model.has_property('deleted'):
            stmt = stmt.filter_by(deleted=False)

        # 处理自定义排序
        if self.order_by and isinstance(self.order_by, (list, tuple)):
            stmt = stmt.order_by(*self.order_by)

        return stmt

    def merge_data(self, data):
        pass

    def get_item(self, **kwargs):
        stmt = self.get_query(**kwargs)
        item = stmt.first()
        if not item:
            raise ZeusNotFound
        return item

    def get_pagination(self, **kwargs):
        stmt = self.get_query(**kwargs)
        page, per_page = self.get_paginate_args()
        pagination = stmt.paginate(page, per_page, error_out=not self.can_empty)
        self.merge_data(pagination.items)
        return pagination

    def get_items(self, **kwargs):
        stmt = self.get_query(**kwargs)
        return stmt.all()


class OutputMixin(object):

    def output_fields(self, fields):
        data = {}
        for name, field in fields.items():
            if isinstance(field, List):
                try:
                    data[name] = [self.output_fields(field.container.nested)]
                except:
                    data[name] = [getattr(field.container, 'prompt', '')]
            elif isinstance(field, Nested):
                data[name] = self.output_fields(field.nested)
            else:
                data[name] = getattr(field, 'prompt', '')

        return data

    def output_args(self, url_parse):
        args = []
        for item in url_parse.args:
            item_data = dict()
            item_data['name'] = item.name
            item_data['help'] = item.help
            item_data['type'] = item.type.__name__
            item_data['required'] = item.required

            if item.choices:
                item_data['choices'] = item.choices

            if item.default:
                item_data['default'] = item.default

            args.append(item_data)
        return args


class ContextMixin(object):
    __context = None

    @property
    def context(self):
        if not self.__context:
            self.__context = dict()
        return self.__context

    def append_context(self, k, v):
        self.context[k] = v

    def remove_context(self, k):
        if k in self.context.keys():
            del self.context[k]


class FormMixin(object):
    # 是否开始csrf验证
    # type: bool
    csrf_enabled = True

    # 创建数据使用的表单
    # type: wtforms obj
    create_form = None

    # 更新数据使用的表单
    # type: wtforms obj
    update_form = None

    # 删除数据使用的表单
    # type: wtforms obj
    delete_form = None

    def get_form(self, form_cls, obj=None, **kwargs):
        form = form_cls(obj=obj, csrf_enabled=self.csrf_enabled)
        for k, v in kwargs.items():
            if form.has_field(k):
                getattr(form, k).data = v
        return form

    def get_create_form(self, **kwargs):
        if not self.create_form:
            raise AttributeError('需要设置create_form')
        return self.get_form(self.create_form, **kwargs)

    def get_update_form(self, **kwargs):
        if not self.update_form:
            raise AttributeError('需要设置update_form')
        return self.get_form(self.update_form, **kwargs)

    def get_delete_form(self, **kwargs):
        if self.delete_form:
            return self.get_form(self.delete_form, **kwargs)


from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.dialects.postgres import UUID
from flask_login import current_user
from .base import (db)
from .crud import (CRUDMixin)
from ..decorators import classproperty
import re


class MarshalLabelMixin:

    def __marshallable__(self):
        data = self.__dict__
        data.update({
            'entry_type': getattr(self, 'cls_entry_type'),
            'entry_id': getattr(self, 'id')
        })
        return data

    @classproperty
    def cls_entry_type(cls):
        regex = re.compile(r'([A-Z])+')
        return regex.sub(lambda x: '/{}'.format(x.group(0).lower()), cls.__name__).strip('/')


class EntryMixin(CRUDMixin):
    # 属性名称
    # 例子: is_followed, is_praised
    attribute = None

    @declared_attr
    def entry_type(self):
        return db.Column('entry_type', db.String(255), index=True)

    @declared_attr
    def entry_id(self):
        return db.Column('entry_id', UUID(as_uuid=True), index=True)

    @classmethod
    def for_entries(cls, items, attribute=None, child=None):
        attribute = attribute or cls.attribute

        if not attribute:
            raise AttributeError('需要设置field值')

        data = items

        if child:
            data = [getattr(item, child) for item in items]

        data_id = set([item.id for item in data])
        results = dict((item_id, False) for item_id in data_id)

        if current_user.is_authenticated:
            query = getattr(cls, 'query')

            res = query\
                .filter(cls.entry_id.in_(data_id))\
                .filter_by(user_id=current_user.id)\
                .all()

            for item in res:
                results[item.entry_id] = True

        for item in data:
            setattr(item, attribute, results.get(item.id))

        return items





from flask import (request, url_for, render_template)
from flask.views import View
from ..common.mixin import QueryMixin, ContextMixin
import types


class BaseView(View, QueryMixin, ContextMixin):
    # 模版
    template = None

    def get_template(self):
        """ 获取模版 """
        if not self.template:
            raise AttributeError('需要设置template值')
        return self.template

    def before_render(self):
        pass

    def after_render(self):
        pass

    def render(self, template=None, context=None):
        """ 渲染模版 """
        template = template or self.get_template()

        if context and isinstance(context, types.MappingProxyType):
            for k, v in context.items():
                self.append_context(k, v)

        self.before_render()
        response = render_template(template, **self.context)
        self.after_render()
        return response

    def get_next_url(self, **kwargs):
        """ 获取跳转链接 """
        return request.args.get('next') or url_for(request.endpoint, **kwargs)




from functools import wraps
from flask import request, current_app


def jsonp(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        callback = request.args.get('callback', False)
        if callback:
            data = str(f(*args, **kwargs))
            content = str(callback) + '(' + data + ')'
            mime_type = 'application/javascript'
            return current_app.response_class(content, mimetype=mime_type)
        else:
            return f(*args, **kwargs)
    return decorated_function


class ClassPropertyDescriptor(object):
    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("Can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



from flask import (request, url_for)
from ..common.mixin import (QueryMixin, OutputMixin, FormMixin)


class BaseResource(QueryMixin, OutputMixin, FormMixin):
    # 输出
    model_fields = None

    # url参数解析
    url_parse = None

    # 是否生成包含域名的完整url
    # type: bool
    is_full_url = True

    @property
    def cls_name(self):
        """ 视图类名称
        :return: str
        """
        return self.__class__.__name__

    def get_model_fields(self):
        """ 检查输出格式是否设置
        :return:
        """
        if not self.model_fields:
            raise AttributeError('{} 需要设置model_fields'.format(self.cls_name))
        return self.model_fields

    def generate_iter_pages(self, pages, per_page, **kwargs):
        """ 生成分页
        :param pages:
        :param per_page:
        :return: list
        """
        pages = list(pages)
        iter_pages = []

        for page in pages:
            if isinstance(page, int):
                iter_pages.append({
                    'page': page,
                    'url': self.generate_url(page, per_page, **kwargs)
                })
            else:
                iter_pages.append({
                    'page': '...',
                    'url': ''
                })

        return iter_pages

    def generate_url(self, page, per_page, **kwargs):
        """ 生成链接
        :param page:
        :param per_page:
        :return: str
        """
        return url_for(request.endpoint, page=page, per_page=per_page, _external=self.is_full_url, **kwargs)



from .base import BaseView


class ListView(BaseView):

    def dispatch_request(self, **kwargs):
        pagination = self.get_pagination(**kwargs)
        self.append_context('items', pagination.items)
        self.append_context('pagination', pagination)
        return self.render()



from .base import BaseView


class DetailView(BaseView):

    def dispatch_request(self, **kwargs):
        item = self.get_item(**kwargs)
        self.append_context('item', item)
        return self.render()



from flask import (flash, redirect)
from flask_login import (login_required, current_user)
from .base import BaseView
from ..common.mixin import FormMixin


class BaseFormView(BaseView, FormMixin):
    success_message = None
    methods = ['GET', 'POST']
    decorators = [login_required]


class CreateView(BaseFormView):

    def dispatch_request(self, **kwargs):
        form = self.get_create_form(**kwargs)

        if form.validate_on_submit():
            model = self.get_model()
            item = model()

            for k, v in form.data.iteritems():
                setattr(item, k, v)

            if model.has_property('user_id'):
                item.user_id = current_user.id

            item.save()
            self.append_context('item', item)

            if self.success_message:
                flash(self.success_message, category='success')

            return redirect(self.get_next_url(**kwargs))

        self.append_context('form', form)
        return self.render(**self.context)


class UpdateView(BaseFormView):

    def dispatch_request(self, **kwargs):

        stmt = self.get_query(**kwargs)
        item = stmt.first_or_404()

        form = self.get_update_form(obj=item, **kwargs)

        if form.validate_on_submit():

            for k, v in form.data.iteritems():
                setattr(item, k, v)

            if item.has_property('user_id'):
                item.user_id = current_user.id

            item.save()
            self.append_context('item', item)

            if self.success_message:
                flash(self.success_message, category='success')

            return redirect(self.get_next_url(**kwargs))

        self.append_context('form', form)
        return self.render(**self.context)


from flask import request
from raven.contrib.flask import Sentry as BaseSentry
from flask_login import current_user


class Sentry(BaseSentry):

    def before_request(self, *args, **kwargs):
        self.last_event_id = None

    def update_context(self):
        if 'request' not in self.client.context:
            self.client.http_context(self.get_http_info(request))
        if 'user' not in self.client.context:
            self.client.user_context(self.get_user_info(request))

    def get_user_info(self, request):
        user_info = super().get_user_info(request)
        if user_info:
            user_info['username'] = current_user.nickname
        return user_info

    def captureException(self, *args, **kwargs):
        self.update_context()
        super().captureException(*args, **kwargs)

    def captureMessage(self, *args, **kwargs):
        self.update_context()
        super().captureMessage(*args, **kwargs)
